//================================================================================================
// $Id: parserext.h 6308 2023-01-16 20:16:41Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Parser extensions
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#ifndef WTC_PARSEREXT_H
#define WTC_PARSEREXT_H

#include "wtparser/frontend.h"
#include <string>
#include <deque>
#include <map>
#include <set>

namespace wtc
{

class SourceManager : public wtparser::ISourceManager
{
public:
	SourceManager();
	virtual wtparser::IFile* OpenFile(const std::string & filename);
};

class ProblemHandler : public wtparser::IProblemHandler
{
public:
	ProblemHandler();

	virtual void AddNote(const std::string & file, unsigned int line, const std::string & msg);
	virtual void AddWarning(const std::string & file, unsigned int line, const std::string & msg);
	virtual void AddError(const std::string & file, unsigned int line, const std::string & msg);

	unsigned long long GetNumNotes()    const { return mNumNotes; }
	unsigned long long GetNumWarnings() const { return mNumWarnings; }
	unsigned long long GetNumErrors()   const { return mNumErrors; }
	unsigned long long GetNumProblems() const { return mNumNotes + mNumWarnings + mNumErrors; }

	struct Problem
	{
		enum Kind
		{
			NOTE,
			WARNING,
			ERROR
		};

		static const std::string & KindToString(Kind kind);

		Problem(Kind kind, const std::string & file, unsigned int line, const std::string & msg)
			: kind(kind)
			, file(file)
			, line(line)
			, msg(msg)
		{
			// ok
		}

		std::string ToString() const;

		const Kind         kind;
		const std::string  file;
		const unsigned int line;
		const std::string  msg;
	};

	typedef std::deque<Problem> problems_t;
	typedef problems_t::iterator iterator;
	typedef problems_t::const_iterator const_iterator;
	iterator begin() { return mProblems.begin(); }
	iterator end() { return mProblems.end(); }
	const_iterator begin() const { return mProblems.begin(); }
	const_iterator end() const { return mProblems.end(); }

private:
	unsigned long long mNumNotes;
	unsigned long long mNumWarnings;
	unsigned long long mNumErrors;
	problems_t         mProblems;
};

class Options : public wtparser::IOptions
{
	std::string mProgramName;
	bool        mDebugLexer;
	bool        mDebugRules;
	bool        mAllowFloatingBlocks;
	bool        mDumpAst;
	bool        mSyntaxOnly;
	bool        mLabelsTop;
	bool        mLabelsAll;
	bool        mInline32;
	bool        mInline64;
	bool        mBigEndian;
	bool        mExpectError;
	bool        mDisableAsserts;
	bool        mDisablePrecond;
	bool        mDisablePostcond;
	bool        mGenerateImportInclude;
	bool        mShebang;
	bool        mSignedIndex;
	bool        mLargeIndex;

	std::set<std::string> mHashAlgs;

	std::string mSourceFilename;
	std::string mOutputFilename;
	std::string mImportFilename;

	typedef void (Options::*Command)(int & argc, char** & argv);

	struct CommandDescriptor
	{
		CommandDescriptor() { }

		CommandDescriptor(const std::string & name, const std::string & help, Command cmd)
			: name(name)
			, help(help)
			, cmd(cmd)
		{
			// ok
		}

		std::string name;
		std::string help;
		Command     cmd;
	};

	typedef std::map<std::string, CommandDescriptor> commands_t;
	commands_t mCommands;

	std::string mProblemString;

	void AddCommand(CommandDescriptor desc);
	void AddCommands();

	static std::string ChangeWTExtensionTo(const std::string & str, const std::string & ext);

public:

	Options(int argc, char** argv);
	virtual bool DebugLexer() const { return mDebugLexer; }
	virtual bool DebugRules() const { return mDebugRules; }
	virtual bool AllowFloatingBlocks() const { return mAllowFloatingBlocks; }

	bool HasProblem() const { return !mProblemString.empty(); }
	const std::string & GetProblem() const { return mProblemString; }

	bool IsHostBigEndian() const;
	bool DumpAst() const { return mDumpAst; }
	bool SyntaxOnly() const { return mSyntaxOnly; }
	bool LabelsTop() const { return mLabelsTop; }
	bool LabelsAll() const { return mLabelsAll; }
	bool BigEndian() const { return mBigEndian; }
	bool Inline32() const { return mInline32; }
	bool Inline64() const { return mInline64; }
	bool ExpectError() const { return mExpectError; }
	bool DisableAsserts() const { return mDisableAsserts; }
	bool EnableAsserts() const { return !mDisableAsserts; }
	bool DisablePrecond() const { return mDisablePrecond; }
	bool EnablePrecond() const { return !mDisablePrecond; }
	bool DisablePostcond() const { return mDisablePostcond; }
	bool EnablePostcond() const { return !mDisablePostcond; }
	bool GenerateImportInclude() const { return mGenerateImportInclude; }
	bool Shebang() const { return mShebang; }
	bool SignedIndex() const { return mSignedIndex; }
	bool LargeIndex() const { return mLargeIndex; }

	const std::set<std::string> & HashAlgs() const { return mHashAlgs; }

	const std::string & GetSourceFilename() const { return mSourceFilename; }
	const std::string & GetOutputFilename() const { return mOutputFilename; }
	const std::string & GetImportFilename() const { return mImportFilename; }

	static const std::string BinaryExtension;
	static const std::string ImportExtension;

private:
	void CmdHelp(int & argc, char** & argv);
	void CmdVersion(int & argc, char** & argv);
	void CmdDumpAst(int & argc, char** & argv);
	void CmdSyntaxOnly(int & argc, char** & argv);
	void CmdDebugLexer(int & argc, char** & argv);
	void CmdDebugRules(int & argc, char** & argv);
	void CmdOutput(int & argc, char** & argv);
	void CmdLabelsNone(int & argc, char** & argv);
	void CmdLabelsTop(int & argc, char** & argv);
	void CmdLabelsAll(int & argc, char** & argv);
	void CmdBigEndian(int & argc, char** & argv);
	void CmdLittleEndian(int & argc, char** & argv);
	void CmdInline32(int & argc, char** & argv);
	void CmdInline64(int & argc, char** & argv);
	void CmdNoInline32(int & argc, char** & argv);
	void CmdNoInline64(int & argc, char** & argv);
	void CmdExpectError(int & argc, char** & argv);
	void CmdDisableAsserts(int & argc, char** & argv);
	void CmdDisablePrecond(int & argc, char** & argv);
	void CmdDisablePostcond(int & argc, char** & argv);
	void CmdGenerateImportInclude(int & argc, char** & argv);
	void CmdImportIncludeName(int & argc, char** & argv);
	void CmdShebang(int & argc, char** & argv);
	void CmdAddHash(int & argc, char** & argv);
	void CmdSignedIndex(int & argc, char** & argv);
	void CmdNoSignedIndex(int & argc, char** & argv);
	void CmdLargeIndex(int & argc, char** & argv);
	void CmdNoLargeIndex(int & argc, char** & argv);
};

} // wtc

#endif // WTC_PARSEREXT_H
