/* $Id: ast_simultaneousexpression.cpp 6310 2023-01-16 20:39:48Z aron $ */
/*
    Welltype parser and AST
    Copyright (C) 2014-2023  Aron Barath

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "ast_simultaneousexpression.h"
#include "ast_simultaneoustype.h"
#include "ast_type.h"
#include "../astvisitor.h"

wtast::SimultaneousExpression::SimultaneousExpression(Expression* expr1, Expression* expr2)
	: Expression()
	, mTypeInfo(new wtast::SimultaneousType())
{
	if(expr1) { AddExpression(expr1); }
	if(expr2) { AddExpression(expr2); }
}

wtast::SimultaneousExpression::~SimultaneousExpression()
{
	while(!mExpressions.empty())
	{
		delete mExpressions.front();
		mExpressions.pop_front();
	}

	delete mTypeInfo;
}

bool wtast::SimultaneousExpression::OriginateFrom(Variable::Class varclass)
{
	for(expressions_t::iterator it=mExpressions.begin(),end=mExpressions.end();it!=end;++it)
	{
		if((*it)->OriginateFrom(varclass))
		{
			return true;
		}
	}

	return Variable::READONLY==varclass || Variable::LOCAL==varclass;
}

void wtast::SimultaneousExpression::AddExpression(Expression* expr)
{
	if(expr)
	{
		if(!expr->IsSimultaneous())
		{
			wtast::TypeInfo* typeinfo = expr->GetType();

			for(std::size_t i=0,n=typeinfo->GetNumberOfTypes();i!=n;++i)
			{
				mTypeInfo->AddType(typeinfo->GetType(i)->Copy());
			}

			mExpressions.push_back(expr);
		}
		else
		{
			SimultaneousExpression* simexpr = static_cast<SimultaneousExpression*>(expr);

			for(expressions_t::iterator it=simexpr->mExpressions.begin(),end=simexpr->mExpressions.end();it!=end;++it)
			{
				AddExpression(*it);
			}
		}
	}
}

wtast::TypeInfo* wtast::SimultaneousExpression::GetType()
{
	return mTypeInfo;
}

bool wtast::SimultaneousExpression::IsTemporary() const
{
	for(expressions_t::const_iterator it=mExpressions.begin(),end=mExpressions.end();it!=end;++it)
	{
		if((*it)->IsTemporary())
		{
			return true;
		}
	}

	return false;
}

void wtast::SimultaneousExpression::Accept(AstVisitor* vis)
{
	if(vis->Visit(this))
	{
		for(expressions_t::iterator it=mExpressions.begin(),end=mExpressions.end();it!=end;++it)
		{
			(*it)->Accept(vis);
		}

		vis->EndVisit(this);
	}
}

wtast::SimultaneousExpression* wtast::SimultaneousExpression::Copy()
{
	SimultaneousExpression* sim = new SimultaneousExpression();

	for(expressions_t::iterator it=mExpressions.begin(),end=mExpressions.end();it!=end;++it)
	{
		sim->AddExpression((*it)->Copy());
	}

	return sim;
}
