/*
   BLAKE2 reference source code package - reference C implementations

   Copyright 2012, Samuel Neves <sneves@dei.uc.pt>.  You may use this under the
   terms of the CC0, the OpenSSL Licence, or the Apache Public License 2.0, at
   your option.  The terms of these licenses can be found at:

   - CC0 1.0 Universal : http://creativecommons.org/publicdomain/zero/1.0
   - OpenSSL license   : https://www.openssl.org/source/license.html
   - Apache 2.0        : http://www.apache.org/licenses/LICENSE-2.0

   More information about the BLAKE2 hash function can be found at
   https://blake2.net.
*/
#ifndef BLAKE2_IMPL_H
#define BLAKE2_IMPL_H

#include "../wtruntime.h"
#include <string.h>

#if !defined(__cplusplus) && (!defined(__STDC_VERSION__) || __STDC_VERSION__ < 199901L)
  #if   defined(_MSC_VER)
    #define BLAKE2_INLINE __inline
  #elif defined(__GNUC__)
    #define BLAKE2_INLINE __inline__
  #else
    #define BLAKE2_INLINE
  #endif
#else
  #define BLAKE2_INLINE inline
#endif

static BLAKE2_INLINE wt_u32 load32( const void *src )
{
#if defined(NATIVE_LITTLE_ENDIAN)
  wt_u32 w;
  memcpy(&w, src, sizeof w);
  return w;
#else
  const wt_u8 *p = ( const wt_u8 * )src;
  return (( wt_u32 )( p[0] ) <<  0) |
         (( wt_u32 )( p[1] ) <<  8) |
         (( wt_u32 )( p[2] ) << 16) |
         (( wt_u32 )( p[3] ) << 24) ;
#endif
}

static BLAKE2_INLINE uint64_t load64( const void *src )
{
#if defined(NATIVE_LITTLE_ENDIAN)
  uint64_t w;
  memcpy(&w, src, sizeof w);
  return w;
#else
  const wt_u8 *p = ( const wt_u8 * )src;
  return (( uint64_t )( p[0] ) <<  0) |
         (( uint64_t )( p[1] ) <<  8) |
         (( uint64_t )( p[2] ) << 16) |
         (( uint64_t )( p[3] ) << 24) |
         (( uint64_t )( p[4] ) << 32) |
         (( uint64_t )( p[5] ) << 40) |
         (( uint64_t )( p[6] ) << 48) |
         (( uint64_t )( p[7] ) << 56) ;
#endif
}

static BLAKE2_INLINE wt_u16 load16( const void *src )
{
#if defined(NATIVE_LITTLE_ENDIAN)
  wt_u16 w;
  memcpy(&w, src, sizeof w);
  return w;
#else
  const wt_u8 *p = ( const wt_u8 * )src;
  return (( wt_u16 )( p[0] ) <<  0) |
         (( wt_u16 )( p[1] ) <<  8) ;
#endif
}

static BLAKE2_INLINE void store16( void *dst, wt_u16 w )
{
#if defined(NATIVE_LITTLE_ENDIAN)
  memcpy(dst, &w, sizeof w);
#else
  wt_u8 *p = ( wt_u8 * )dst;
  *p++ = ( wt_u8 )w; w >>= 8;
  *p++ = ( wt_u8 )w;
#endif
}

static BLAKE2_INLINE void store32( void *dst, wt_u32 w )
{
#if defined(NATIVE_LITTLE_ENDIAN)
  memcpy(dst, &w, sizeof w);
#else
  wt_u8 *p = ( wt_u8 * )dst;
  p[0] = (wt_u8)(w >>  0);
  p[1] = (wt_u8)(w >>  8);
  p[2] = (wt_u8)(w >> 16);
  p[3] = (wt_u8)(w >> 24);
#endif
}

static BLAKE2_INLINE void store64( void *dst, uint64_t w )
{
#if defined(NATIVE_LITTLE_ENDIAN)
  memcpy(dst, &w, sizeof w);
#else
  wt_u8 *p = ( wt_u8 * )dst;
  p[0] = (wt_u8)(w >>  0);
  p[1] = (wt_u8)(w >>  8);
  p[2] = (wt_u8)(w >> 16);
  p[3] = (wt_u8)(w >> 24);
  p[4] = (wt_u8)(w >> 32);
  p[5] = (wt_u8)(w >> 40);
  p[6] = (wt_u8)(w >> 48);
  p[7] = (wt_u8)(w >> 56);
#endif
}

static BLAKE2_INLINE uint64_t load48( const void *src )
{
  const wt_u8 *p = ( const wt_u8 * )src;
  return (( uint64_t )( p[0] ) <<  0) |
         (( uint64_t )( p[1] ) <<  8) |
         (( uint64_t )( p[2] ) << 16) |
         (( uint64_t )( p[3] ) << 24) |
         (( uint64_t )( p[4] ) << 32) |
         (( uint64_t )( p[5] ) << 40) ;
}

static BLAKE2_INLINE void store48( void *dst, uint64_t w )
{
  wt_u8 *p = ( wt_u8 * )dst;
  p[0] = (wt_u8)(w >>  0);
  p[1] = (wt_u8)(w >>  8);
  p[2] = (wt_u8)(w >> 16);
  p[3] = (wt_u8)(w >> 24);
  p[4] = (wt_u8)(w >> 32);
  p[5] = (wt_u8)(w >> 40);
}

static BLAKE2_INLINE wt_u32 rotr32( const wt_u32 w, const unsigned c )
{
  return ( w >> c ) | ( w << ( 32 - c ) );
}

static BLAKE2_INLINE uint64_t rotr64( const uint64_t w, const unsigned c )
{
  return ( w >> c ) | ( w << ( 64 - c ) );
}

/* prevents compiler optimizing out memset() */
static BLAKE2_INLINE void secure_zero_memory(void *v, size_t n)
{
  static void *(*const volatile memset_v)(void *, int, size_t) = &memset;
  memset_v(v, 0, n);
}

#endif
